'use strict';

/*
  tests are going to be really small, basically make sure
  they create the standard stuff and capture the correct stack trace
*/

const express = require('express');
const axios = require('axios');
const { expect } = require('chai');

const errors = require('../index.js');

context('All Tests', () => {
  describe('Axios', () => {
    const app = express();
    const listenPort = 10191; // shouldn't conflict with potential containers running

    let server;

    before(done => {
      app.get('/returns/400', (req, res) => res.status(400).send({ message: 'some message ya' }));
      app.get('/returns/404', (req, res) => res.status(404).send({ message: 'not found' }));

      server = app.listen(listenPort, done);
    });

    after(done => {
      server.close(done);
    });

    it('Handles ECONNREFUSED', async () => {
      try {
        await axios({
          method: 'get',
          url: '/this/is/not/a/responsive/url'
        });
      } catch (error) {
        const { stack: errStack } = new Error();
        const match = errStack.match(/all\.test\.js:(\d+):\d+\)/);
        const lineNumber = Number(match[1]) + 5; // add to match the line number of the errors.Axios()
        const {
          name, stack, statusCode, errorCode, message
        } = new errors.Axios(error);
        expect(name).to.equal('Axios');
        expect(message).to.equal('Invalid URL');
        expect(stack).to.match(/^Axios: Invalid URL/);
        expect(stack).to.match(new RegExp(`/test/all\\.test\\.js:${lineNumber}:\\d+\\)`));
        expect(statusCode).to.equal(500);
        expect(errorCode).to.equal('AXIOS_ERR_INVALID_URL');
        return;
      }

      expect.fail('Axios did not throw an error, try changing the url to somewhere that refuses to connect');
    });

    it('Handles a 400 response', async () => {
      try {
        await axios({
          method: 'get',
          url: `http://localhost:${listenPort}/returns/400`
        });
      } catch (error) {
        const { stack: errStack } = new Error();
        const match = errStack.match(/all\.test\.js:(\d+):\d+\)\n/);
        const lineNumber = Number(match[1]) + 5; // add to match the line number of the errors.Axios()
        const {
          name, stack, statusCode, errorCode, message
        } = new errors.Axios(error);
        expect(name).to.equal('Axios');
        expect(message).to.equal('Request failed with status code 400');
        expect(stack).to.match(/^Axios: Request failed with status code 400/);
        expect(stack).to.match(new RegExp(`/test/all\\.test\\.js:${lineNumber}:\\d+\\)`));
        expect(statusCode).to.equal(400);
        expect(errorCode).to.equal('AXIOS_ERR_BAD_REQUEST');
        return;
      }

      expect.fail('Axios should have thrown an error, check your express settings and axios url');
    });

    it('Handles a 404 response', async () => {
      try {
        await axios({
          method: 'get',
          url: `http://localhost:${listenPort}/returns/404`
        });
      } catch (error) {
        const { stack: errStack } = new Error();
        const match = errStack.match(/all\.test\.js:(\d+):\d+\)\n/);
        const lineNumber = Number(match[1]) + 5; // add to match the line number of the errors.Axios()
        const {
          name, stack, statusCode, errorCode, message
        } = new errors.Axios(error);
        expect(name).to.equal('Axios');
        expect(message).to.equal('Request failed with status code 404');
        expect(stack).to.match(/^Axios: Request failed with status code 404/);
        expect(stack).to.match(new RegExp(`/test/all\\.test\\.js:${lineNumber}:\\d+\\)`));
        expect(statusCode).to.equal(404);
        expect(errorCode).to.equal('AXIOS_ERR_BAD_REQUEST');
        return;
      }

      expect.fail('Axios should have thrown an error, check your express settings and axios url');
    });

    it('Takes custom messages and errorCodes', async () => {
      try {
        await axios({
          method: 'get',
          url: `http://localhost:${listenPort}/returns/404`
        });
      } catch (error) {
        const { stack: errStack } = new Error();
        const match = errStack.match(/all\.test\.js:(\d+):\d+\)\n/);
        const lineNumber = Number(match[1]) + 5; // add to match the line number of the errors.Axios()
        const {
          name, stack, statusCode, errorCode, message
        } = new errors.Axios(error, { message: 'this is a custom message', customErrorCode: 'this_is a BAD error\ncode' });
        expect(name).to.equal('Axios');
        expect(message).to.equal('this is a custom message');
        expect(stack).to.match(/^Axios: this is a custom message/);
        expect(stack).to.match(new RegExp(`/test/all\\.test\\.js:${lineNumber}:\\d+\\)`));
        expect(statusCode).to.equal(404);
        expect(errorCode).to.equal('AXIOS_THIS_IS_A_BAD_ERROR_CODE');
        return;
      }

      expect.fail('Axios should have thrown an error, check your express settings and axios url');
    });

    it('Ignores non-object options parameter', () => {
      const { errorCode, message } = new errors.Axios({}, 'not an object');
      expect(errorCode).to.equal('AXIOS_GENERIC');
      expect(message).to.equal('An error has occured');
    });
  });

  describe('Forbidden', () => {
    it('works', () => {
      const { stack: errStack } = new Error();
      const match = errStack.match(/all\.test\.js:(\d+):\d+\)\n/);
      const lineNumber = Number(match[1]) + 5; // add to match the line number of the error.Forbidden
      const {
        name, stack, statusCode, errorCode, message
      } = new errors.Forbidden('some message', 'this is an error');
      expect(name).to.equal('Forbidden');
      expect(message).to.equal('some message');
      expect(stack).to.match(/^Forbidden: some message/);
      expect(stack).to.match(new RegExp(`/test/all\\.test\\.js:${lineNumber}:\\d+\\)`));
      expect(statusCode).to.equal(403);
      expect(errorCode).to.equal('THIS_IS_AN_ERROR');
    });

    it('defaults the errorCode to FORBIDDEN', () => {
      const { errorCode } = new errors.Forbidden('an error');
      expect(errorCode).to.equal('FORBIDDEN');
    });
  });

  describe('InvalidInput', () => {
    it('works', () => {
      const { stack: errStack } = new Error();
      const match = errStack.match(/all\.test\.js:(\d+):\d+\)\n/);
      const lineNumber = Number(match[1]) + 5; // add to match the line number of the error.InvalidInput
      const {
        name, stack, statusCode, errorCode, message, dataPath
      } = new errors.InvalidInput('some message', { dataPath: '.this.path.here' });
      expect(name).to.equal('InvalidInput');
      expect(message).to.equal('some message');
      expect(stack).to.match(/^InvalidInput: some message/);
      expect(stack).to.match(new RegExp(`/test/all\\.test\\.js:${lineNumber}:\\d+\\)`));
      expect(statusCode).to.equal(400);
      expect(errorCode).to.equal('INVALID_INPUT');
      expect(dataPath).to.equal('.this.path.here');
    });

    it('ignores non-object options', () => {
      const { stack: errStack } = new Error();
      const match = errStack.match(/all\.test\.js:(\d+):\d+\)\n/);
      const lineNumber = Number(match[1]) + 5; // add to match the line number of the error.InvalidInput
      const {
        name, stack, statusCode, errorCode, message, dataPath
      } = new errors.InvalidInput('some message', 'yo its not an object');
      expect(name).to.equal('InvalidInput');
      expect(message).to.equal('some message');
      expect(stack).to.match(/^InvalidInput: some message/);
      expect(stack).to.match(new RegExp(`/test/all\\.test\\.js:${lineNumber}:\\d+\\)`));
      expect(statusCode).to.equal(400);
      expect(errorCode).to.equal('INVALID_INPUT');
      expect(dataPath).to.equal(undefined);
    });

    it('defaults the errorCode to INVALID_INPUT', () => {
      const { errorCode } = new errors.InvalidInput('an error');
      expect(errorCode).to.equal('INVALID_INPUT');
    });
  });

  describe('ResourceNotFound', () => {
    it('works', () => {
      const { stack: errStack } = new Error();
      const match = errStack.match(/all\.test\.js:(\d+):\d+\)\n/);
      const lineNumber = Number(match[1]) + 5; // add to match the line number of the error.InvalidInput
      const {
        name, stack, statusCode, errorCode, message
      } = new errors.ResourceNotFound('some message', 'whoops cant do that');
      expect(name).to.equal('ResourceNotFound');
      expect(message).to.equal('some message');
      expect(stack).to.match(/^ResourceNotFound: some message/);
      expect(stack).to.match(new RegExp(`/test/all\\.test\\.js:${lineNumber}:\\d+\\)`));
      expect(statusCode).to.equal(404);
      expect(errorCode).to.equal('WHOOPS_CANT_DO_THAT');
    });

    it('defaults the errorCode to NOT_FOUND', () => {
      const { errorCode } = new errors.ResourceNotFound('an error');
      expect(errorCode).to.equal('NOT_FOUND');
    });
  });

  describe('Unauthorized', () => {
    it('works', () => {
      const { stack: errStack } = new Error();
      const match = errStack.match(/all\.test\.js:(\d+):\d+\)\n/);
      const lineNumber = Number(match[1]) + 5; // add to match the line number of the error.InvalidInput
      const {
        name, stack, statusCode, errorCode, message
      } = new errors.Unauthorized('some message', 'whoops cant do that');
      expect(name).to.equal('Unauthorized');
      expect(message).to.equal('some message');
      expect(stack).to.match(/^Unauthorized: some message/);
      expect(stack).to.match(new RegExp(`/test/all\\.test\\.js:${lineNumber}:\\d+\\)`));
      expect(statusCode).to.equal(401);
      expect(errorCode).to.equal('WHOOPS_CANT_DO_THAT');
    });

    it('defaults the errorCode to UNAUTHORIZED', () => {
      const { errorCode } = new errors.Unauthorized('an error');
      expect(errorCode).to.equal('UNAUTHORIZED');
    });
  });

  describe('Conflict', () => {
    it('works', () => {
      const { stack: errStack } = new Error();
      const match = errStack.match(/all\.test\.js:(\d+):\d+\)\n/);
      const lineNumber = Number(match[1]) + 5; // add to match the line number of the error.InvalidInput
      const {
        name, stack, statusCode, errorCode, message
      } = new errors.Conflict('some message', 'whoops cant do that');
      expect(name).to.equal('Conflict');
      expect(message).to.equal('some message');
      expect(stack).to.match(/^Conflict: some message/);
      expect(stack).to.match(new RegExp(`/test/all\\.test\\.js:${lineNumber}:\\d+\\)`));
      expect(statusCode).to.equal(409);
      expect(errorCode).to.equal('WHOOPS_CANT_DO_THAT');
    });

    it('defaults the errorCode to CONFLICT', () => {
      const { errorCode } = new errors.Conflict('an error');
      expect(errorCode).to.equal('CONFLICT');
    });
  });

  describe('Rpc', () => {
    it('takes the status, message, result, errors from response', () => {
      const responseMessage = {
        status: 418,
        message: 'im a teapot',
        result: { the: 'object' },
        errors: [1, 2, 3]
      };

      const { stack: errStack } = new Error();
      const match = errStack.match(/all\.test\.js:(\d+):\d+\)\n/);
      const lineNumber = Number(match[1]) + 5; // add to match the line number of the errors.Rpc
      const {
        name, stack, statusCode, errorCode, message, result, errors: errArray
      } = new errors.Rpc(responseMessage);
      expect(name).to.equal('Rpc');
      expect(message).to.equal('im a teapot');
      expect(stack).to.match(/^Rpc: im a teapot/);
      expect(stack).to.match(new RegExp(`/test/all\\.test\\.js:${lineNumber}:\\d+\\)`));
      expect(statusCode).to.equal(418);
      expect(errorCode).to.equal('RPC_GENERIC');
      expect(result).to.equal(responseMessage.result);
      expect(errArray).to.equal(responseMessage.errors);
    });
  });
});
